#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import logging
import platform

from connect.eaas.runner.helpers import (
    get_environment,
    get_version,
)
from connect.eaas.core.proto import Logging, LogMeta

logger = logging.getLogger(__name__)


class ConfigHelper:
    """
    Provides both static (from env vars) and dynamic (from backend)
    configuration.
    """
    def __init__(self, secure=True):
        self.secure = secure
        self.env = get_environment()
        self.dyn_config = None

    @property
    def api_key(self):
        return self.env['api_key']

    @property
    def service_id(self):
        return self.dyn_config.logging.meta.service_id

    @property
    def products(self):
        return self.dyn_config.logging.meta.products

    @property
    def hub_id(self):
        return self.dyn_config.logging.meta.hub_id

    @property
    def environment_id(self):
        return self.env['environment_id']

    @property
    def instance_id(self):
        return self.env['instance_id']

    @property
    def environment_type(self):
        return self.dyn_config.environment_type

    @property
    def account_id(self):
        return self.dyn_config.logging.meta.account_id

    @property
    def account_name(self):
        return self.dyn_config.logging.meta.account_name

    @property
    def logging_api_key(self):
        return self.dyn_config.logging.logging_api_key

    @property
    def logging_level(self):
        return self.dyn_config.logging.log_level

    @property
    def variables(self):
        return self.dyn_config.variables

    @property
    def webapp_port(self):
        return self.env['webapp_port']

    @property
    def metadata(self):
        return {
            'api_address': self.env['api_address'],
            'service_id': self.service_id,
            'environment_id': self.environment_id,
            'environment_type': self.environment_type,
            'instance_id': self.instance_id,
            'account_id': self.account_id,
            'account_name': self.account_name,
        }

    @property
    def event_definitions(self):
        return {
            definition.event_type: definition
            for definition in (self.dyn_config.event_definitions or [])
        }

    def get_events_ws_url(self):
        proto = 'wss' if self.secure else 'ws'
        return (
            f'{proto}://{self.env["ws_address"]}/public/v1/devops/ws'
            f'/{self.env["environment_id"]}/{self.env["instance_id"]}'
        )

    def get_webapp_ws_url(self):
        proto = 'wss' if self.secure else 'ws'
        return (
            f'{proto}://{self.env["ws_address"]}/public/v1/devops/ws'
            f'/{self.env["environment_id"]}/{self.env["instance_id"]}/webapp'
        )

    def get_api_url(self):
        return f'https://{self.env["api_address"]}/public/v1'

    def get_user_agent(self):
        version = get_version()
        pimpl = platform.python_implementation()
        pver = platform.python_version()
        sysname = platform.system()
        sysver = platform.release()
        ua = (
            f'connect-extension-runner/{version} {pimpl}/{pver} {sysname}/{sysver}'
            f' {self.env["environment_id"]}/{self.env["instance_id"]}'
        )
        return {'User-Agent': ua}

    def get_headers(self):
        return (
            ('Authorization', self.api_key),
        )

    def get_timeout(self, category):
        return self.env[f'{category}_task_max_execution_time']

    def update_dynamic_config(self, data):
        """Updates the dynamic configuration."""
        if data.logging is None:
            data.logging = Logging()

        if data.logging.meta is None:
            data.logging.meta = LogMeta()

        if not self.dyn_config:
            self.dyn_config = data
        else:
            self.dyn_config.logging.meta.service_id = (
                data.logging.meta.service_id or self.dyn_config.logging.meta.service_id
            )
            self.dyn_config.logging.meta.products = (
                data.logging.meta.products or self.dyn_config.logging.meta.products
            )
            self.dyn_config.logging.meta.hub_id = (
                data.logging.meta.hub_id or self.dyn_config.logging.meta.hub_id
            )
            self.dyn_config.environment_type = (
                data.environment_type or self.dyn_config.environment_type
            )
            self.dyn_config.logging.meta.account_id = (
                data.logging.meta.account_id or self.dyn_config.logging.meta.account_id
            )
            self.dyn_config.logging.meta.account_name = (
                data.logging.meta.account_name or self.dyn_config.logging.meta.account_name
            )
            self.dyn_config.variables = data.variables or self.dyn_config.variables
            self.dyn_config.logging.logging_api_key = (
                data.logging.logging_api_key or self.dyn_config.logging.logging_api_key
            )

            self.dyn_config.logging.logging_api_key = (
                data.logging.logging_api_key or self.dyn_config.logging.logging_api_key
            )

            self.dyn_config.logging.log_level = (
                data.logging.log_level or self.dyn_config.logging.log_level
            )

        logger.debug(f'Runner dynamic config updated {data}')
        if data.logging.log_level:
            logger.info(f'Change extension logger level to {data.logging.log_level}')
            logging.getLogger('eaas.extension').setLevel(
                getattr(logging, data.logging.log_level),
            )
        if data.logging.runner_log_level:
            logging.getLogger('connect.eaas').setLevel(
                getattr(logging, data.logging.runner_log_level),
            )



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import os
import subprocess
from uuid import uuid4

from pkg_resources import (
    DistributionNotFound,
    get_distribution,
)

from connect.eaas.runner.constants import (
    BACKGROUND_TASK_MAX_EXECUTION_TIME,
    INTERACTIVE_TASK_MAX_EXECUTION_TIME,
    ORDINAL_SUFFIX,
    SCHEDULED_TASK_MAX_EXECUTION_TIME,
)


def get_container_id():
    result = subprocess.run(
        ['cat', '/proc/1/cpuset'],
        capture_output=True,
        stdin=subprocess.DEVNULL,
        start_new_session=True,
    )
    try:
        result.check_returncode()
    except subprocess.CalledProcessError:
        return str(uuid4())

    _, container_id = result.stdout.decode()[:-1].rsplit('/', 1)
    if len(container_id) == 64:
        return container_id

    result = subprocess.run(
        ['grep', 'overlay', '/proc/self/mountinfo'],
        capture_output=True,
        stdin=subprocess.DEVNULL,
        start_new_session=True,
    )
    try:
        result.check_returncode()
        mount = result.stdout.decode()
        start_idx = mount.index('upperdir=') + len('upperdir=')
        end_idx = mount.index(',', start_idx)
        dir_path = mount[start_idx:end_idx]
        _, container_id, _ = dir_path.rsplit('/', 2)
        if len(container_id) != 64:
            return str(uuid4())
        return container_id
    except (subprocess.CalledProcessError, ValueError):
        return str(uuid4())


def get_environment():
    return {
        'api_key': os.getenv('API_KEY'),
        'environment_id': os.getenv('ENVIRONMENT_ID'),
        'instance_id': os.getenv('INSTANCE_ID', get_container_id()),
        'ws_address': os.getenv('SERVER_ADDRESS', 'api.cnct.info'),
        'api_address': os.getenv('API_ADDRESS', os.getenv('SERVER_ADDRESS', 'api.cnct.info')),
        'webapp_port': int(os.getenv('WEBAPP_PORT', '53537')),
        'background_task_max_execution_time': int(os.getenv(
            'BACKGROUND_TASK_MAX_EXECUTION_TIME', BACKGROUND_TASK_MAX_EXECUTION_TIME,
        )),
        'interactive_task_max_execution_time': int(os.getenv(
            'INTERACTIVE_TASK_MAX_EXECUTION_TIME', INTERACTIVE_TASK_MAX_EXECUTION_TIME,
        )),
        'scheduled_task_max_execution_time': int(os.getenv(
            'SCHEDULED_TASK_MAX_EXECUTION_TIME', SCHEDULED_TASK_MAX_EXECUTION_TIME,
        )),
    }


def get_version():
    try:
        return get_distribution('connect-extension-runner').version
    except DistributionNotFound:
        return '0.0.0'


def to_ordinal(val):
    if val > 14:
        return f"{val}{ORDINAL_SUFFIX.get(int(str(val)[-1]), 'th')}"
    return f"{val}{ORDINAL_SUFFIX.get(val, 'th')}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['connect',
 'connect.eaas',
 'connect.eaas.runner',
 'connect.eaas.runner.managers']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=1.11.1,<2.0.0',
 'connect-eaas-core>=2.3.0,<3',
 'connect-openapi-client>=25.4',
 'httpx>=0.23.0,<0.24.0',
 'logzio-python-handler>=3.0.0,<4.0.0',
 'uvicorn>=0.17.6,<0.18.0',
 'uvloop>=0.16.0,<0.17.0',
 'websockets>=10.0.0,<11.0.0']

entry_points = \
{'console_scripts': ['cextrun = connect.eaas.runner.main:main']}

setup_kwargs = {
    'name': 'connect-extension-runner',
    'version': '26.5',
    'description': 'CloudBlue Connect EaaS Extension Runner',
    'long_description': '# CloudBlue Connect EaaS Extension Runner\n\n\n![pyversions](https://img.shields.io/pypi/pyversions/connect-extension-runner.svg) [![PyPi Status](https://img.shields.io/pypi/v/connect-extension-runner.svg)](https://pypi.org/project/connect-extension-runner/) [![Build Connect Reports Core](https://github.com/cloudblue/connect-extension-runner/actions/workflows/build.yml/badge.svg)](https://github.com/cloudblue/connect-extension-runner/actions/workflows/build.yml) [![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=connect-extension-runner&metric=alert_status)](https://sonarcloud.io/dashboard?id=connect-extension-runner) [![Coverage](https://sonarcloud.io/api/project_badges/measure?project=connect-extension-runner&metric=coverage)](https://sonarcloud.io/dashboard?id=connect-extension-runner) [![Maintainability Rating](https://sonarcloud.io/api/project_badges/measure?project=connect-extension-runner&metric=sqale_rating)](https://sonarcloud.io/dashboard?id=connect-extension-runner)![Docker Image Version (latest semver)](https://img.shields.io/docker/v/cloudblueconnect/connect-extension-runner?label=docker%20version&sort=semver)![Docker Image Size (latest semver)](https://img.shields.io/docker/image-size/cloudblueconnect/connect-extension-runner?label=docker%20image%20size&sort=semver)![Docker Pulls](https://img.shields.io/docker/pulls/cloudblueconnect/connect-extension-runner)\n',
    'author': 'CloudBlue LLC',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://connect.cloudblue.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#

from connect.eaas.core.enums import EventType


EVENT_TYPE_EXT_METHOD_MAP = {
    EventType.ASSET_PURCHASE_REQUEST_PROCESSING: 'process_asset_purchase_request',
    EventType.ASSET_CHANGE_REQUEST_PROCESSING: 'process_asset_change_request',
    EventType.ASSET_SUSPEND_REQUEST_PROCESSING: 'process_asset_suspend_request',
    EventType.ASSET_RESUME_REQUEST_PROCESSING: 'process_asset_resume_request',
    EventType.ASSET_CANCEL_REQUEST_PROCESSING: 'process_asset_cancel_request',
    EventType.ASSET_ADJUSTMENT_REQUEST_PROCESSING: 'process_asset_adjustment_request',
    EventType.ASSET_PURCHASE_REQUEST_VALIDATION: 'validate_asset_purchase_request',
    EventType.ASSET_CHANGE_REQUEST_VALIDATION: 'validate_asset_change_request',
    EventType.PRODUCT_ACTION_EXECUTION: 'execute_product_action',
    EventType.PRODUCT_CUSTOM_EVENT_PROCESSING: 'process_product_custom_event',
    EventType.TIER_CONFIG_SETUP_REQUEST_PROCESSING: 'process_tier_config_setup_request',
    EventType.TIER_CONFIG_CHANGE_REQUEST_PROCESSING: 'process_tier_config_change_request',
    EventType.TIER_CONFIG_ADJUSTMENT_REQUEST_PROCESSING: 'process_tier_config_adjustment_request',
    EventType.TIER_CONFIG_SETUP_REQUEST_VALIDATION: 'validate_tier_config_setup_request',
    EventType.TIER_CONFIG_CHANGE_REQUEST_VALIDATION: 'validate_tier_config_change_request',
    EventType.LISTING_NEW_REQUEST_PROCESSING: 'process_new_listing_request',
    EventType.LISTING_REMOVE_REQUEST_PROCESSING: 'process_remove_listing_request',
    EventType.TIER_ACCOUNT_UPDATE_REQUEST_PROCESSING: 'process_tier_account_update_request',
    EventType.USAGE_FILE_REQUEST_PROCESSING: 'process_usage_file',
    EventType.PART_USAGE_FILE_REQUEST_PROCESSING: 'process_usage_chunk_file',
}

ASSET_REQUEST_EVENT_TYPES = (
    EventType.ASSET_PURCHASE_REQUEST_PROCESSING,
    EventType.ASSET_CHANGE_REQUEST_PROCESSING,
    EventType.ASSET_SUSPEND_REQUEST_PROCESSING,
    EventType.ASSET_RESUME_REQUEST_PROCESSING,
    EventType.ASSET_CANCEL_REQUEST_PROCESSING,
    EventType.ASSET_ADJUSTMENT_REQUEST_PROCESSING,
    EventType.ASSET_PURCHASE_REQUEST_VALIDATION,
    EventType.ASSET_CHANGE_REQUEST_VALIDATION,
)

TIER_CONFIG_REQUEST_EVENT_TYPES = (
    EventType.TIER_CONFIG_SETUP_REQUEST_PROCESSING,
    EventType.TIER_CONFIG_CHANGE_REQUEST_PROCESSING,
    EventType.TIER_CONFIG_SETUP_REQUEST_VALIDATION,
    EventType.TIER_CONFIG_CHANGE_REQUEST_VALIDATION,
    EventType.TIER_CONFIG_ADJUSTMENT_REQUEST_PROCESSING,
)

LISTING_REQUEST_EVENT_TYPES = (
    EventType.LISTING_NEW_REQUEST_PROCESSING,
    EventType.LISTING_REMOVE_REQUEST_PROCESSING,
)

BACKGROUND_EVENT_TYPES = (
    EventType.ASSET_PURCHASE_REQUEST_PROCESSING,
    EventType.ASSET_CHANGE_REQUEST_PROCESSING,
    EventType.ASSET_SUSPEND_REQUEST_PROCESSING,
    EventType.ASSET_RESUME_REQUEST_PROCESSING,
    EventType.ASSET_CANCEL_REQUEST_PROCESSING,
    EventType.ASSET_ADJUSTMENT_REQUEST_PROCESSING,
    EventType.TIER_CONFIG_SETUP_REQUEST_PROCESSING,
    EventType.TIER_CONFIG_CHANGE_REQUEST_PROCESSING,
    EventType.TIER_CONFIG_ADJUSTMENT_REQUEST_PROCESSING,
    EventType.LISTING_NEW_REQUEST_PROCESSING,
    EventType.LISTING_REMOVE_REQUEST_PROCESSING,
    EventType.TIER_ACCOUNT_UPDATE_REQUEST_PROCESSING,
    EventType.USAGE_FILE_REQUEST_PROCESSING,
    EventType.PART_USAGE_FILE_REQUEST_PROCESSING,

)

INTERACTIVE_EVENT_TYPES = (
    EventType.ASSET_PURCHASE_REQUEST_VALIDATION,
    EventType.ASSET_CHANGE_REQUEST_VALIDATION,
    EventType.TIER_CONFIG_SETUP_REQUEST_VALIDATION,
    EventType.TIER_CONFIG_CHANGE_REQUEST_VALIDATION,
    EventType.PRODUCT_ACTION_EXECUTION,
    EventType.PRODUCT_CUSTOM_EVENT_PROCESSING,
)

VALIDATION_EVENT_TYPES = (
    EventType.ASSET_PURCHASE_REQUEST_VALIDATION,
    EventType.ASSET_CHANGE_REQUEST_VALIDATION,
    EventType.TIER_CONFIG_SETUP_REQUEST_VALIDATION,
    EventType.TIER_CONFIG_CHANGE_REQUEST_VALIDATION,
)

OTHER_INTERACTIVE_EVENT_TYPES = (
    EventType.PRODUCT_ACTION_EXECUTION,
    EventType.PRODUCT_CUSTOM_EVENT_PROCESSING,
)

BACKGROUND_TASK_MAX_EXECUTION_TIME = 300
INTERACTIVE_TASK_MAX_EXECUTION_TIME = 120
SCHEDULED_TASK_MAX_EXECUTION_TIME = 60 * 60 * 12
RESULT_SENDER_MAX_RETRIES = 5
RESULT_SENDER_WAIT_GRACE_SECONDS = 90

MAX_RETRY_TIME_GENERIC_SECONDS = 15 * 60
MAX_RETRY_TIME_MAINTENANCE_SECONDS = 3 * 60 * 60
MAX_RETRY_DELAY_TIME_SECONDS = 5 * 60

DELAY_ON_CONNECT_EXCEPTION_SECONDS = 5

ORDINAL_SUFFIX = {
    1: 'st',
    2: 'nd',
    3: 'rd',
    11: 'th',
    12: 'th',
    13: 'th',
}



import inspect
import logging

from pkg_resources import iter_entry_points

from connect.client import AsyncConnectClient, ConnectClient
from connect.eaas.core.logging import ExtensionLogHandler, RequestLogger
from connect.eaas.runner.config import ConfigHelper
from connect.eaas.runner.constants import EVENT_TYPE_EXT_METHOD_MAP


logger = logging.getLogger(__name__)


class ExtensionHandler:
    """
    Handle the lifecycle of an extension.
    """
    def __init__(self, config: ConfigHelper):
        self._config = config
        self._extension_class = self.get_extension_class()
        if self._extension_class:
            self._descriptor = self._extension_class.get_descriptor()
            self._events = self.get_events()
            self._schedulables = self.get_schedulables()
            self._variables = self.get_variables()
        self._logging_handler = None

    @property
    def should_start(self):
        return self._extension_class is not None

    @property
    def events(self):
        return self._events

    @property
    def variables(self):
        return self._variables

    @property
    def schedulables(self):
        return self._schedulables

    @property
    def readme(self):
        return self._descriptor['readme_url']

    @property
    def changelog(self):
        return self._descriptor['changelog_url']

    def get_method(self, task_id, method_name, installation=None, api_key=None):
        if not method_name:  # pragma: no cover
            return
        args = (
            self._create_client(task_id, method_name, self._config.api_key),
            self.get_logger(task_id),
            self._config.variables,
        )
        kwargs = {}
        if installation:
            kwargs['installation'] = installation
            kwargs['installation_client'] = self._create_client(task_id, method_name, api_key)

        ext = self._extension_class(*args, **kwargs)

        return getattr(ext, method_name, None)

    def get_logger(self, task_id):
        """
        Returns a logger instance configured with the LogZ.io handler.
        This logger will be used by the extension to send logging records
        to the Logz.io service.
        """
        logger = logging.getLogger('eaas.extension')
        if self._logging_handler is None and self._config.logging_api_key is not None:
            self._logging_handler = ExtensionLogHandler(
                self._config.logging_api_key,
                default_extra_fields=self._config.metadata,
            )
            logger.addHandler(self._logging_handler)
        return logging.LoggerAdapter(
            logger,
            {'task_id': task_id},
        )

    def _create_client(self, task_id, method_name, api_key):
        """
        Get an instance of the Connect Openapi Client. If the extension is asyncrhonous
        it returns an instance of the AsyncConnectClient otherwise the ConnectClient.
        """
        method = getattr(self._extension_class, method_name)

        Client = ConnectClient if not inspect.iscoroutinefunction(method) else AsyncConnectClient

        return Client(
            api_key,
            endpoint=self._config.get_api_url(),
            use_specs=False,
            logger=RequestLogger(
                self.get_logger(task_id),
            ),
            default_headers=self._config.get_user_agent(),
        )

    def get_events(self):
        if 'capabilities' in self._descriptor:
            logger.warning(
                "The definition of extension's capabilities in extension.json is deprecated.",
            )
            data = {
                event_type: {
                    'method': EVENT_TYPE_EXT_METHOD_MAP[event_type],
                    'event_type': event_type,
                    'statuses': statuses,
                }
                for event_type, statuses in self._descriptor['capabilities'].items()
            }
            return data

        return {
            event['event_type']: event
            for event in self._extension_class.get_events()
        }

    def get_schedulables(self):
        if 'schedulables' in self._descriptor:
            logger.warning(
                "The definition of extension's schedulables in extension.json is deprecated.",
            )
            return self._descriptor['schedulables']
        return self._extension_class.get_schedulables()

    def get_variables(self):
        if 'variables' in self._descriptor:
            logger.warning(
                "The definition of extension's variables in extension.json is deprecated.",
            )
            return self._descriptor['variables']
        return self._extension_class.get_variables()

    def get_extension_class(self):
        ext_class = next(iter_entry_points('connect.eaas.ext', 'extension'), None)
        return ext_class.load() if ext_class else None
